/*============================================================================
  File:     The Tipping Point.sql

  Summary:  Various examples and ways to see the estimated and actual tipping
			point of when a nonclustered index is not used vs. doing a table
			scan.
  
  Date:     November 2010

  SQL Server Version: SQL Server 2005/2008
------------------------------------------------------------------------------
  Written by Kimberly L. Tripp, SYSolutions, Inc.

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  given by Kimberly L. Tripp.  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

USE AdventureWorksDW2008
go

sp_helpindex FactInternetSales
go

SELECT * FROM sys.dm_db_index_physical_stats
	(db_id(), object_id('FactInternetSales'), NULL, NULL, 'limited')

-- Clustered Index has 1025 PAGES
-- 1/4 = ~256
------------
-- 1/3 = ~338

SELECT S.CustomerKey, COUNT(*) AS NumTotalSales
FROM FactInternetSales AS S
GROUP BY S.CustomerKey
ORDER BY S.CustomerKey 
COMPUTE AVG(COUNT(*))
go

-- Turn on showplan (Ctrl+K)
SET STATISTICS IO ON
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11048
GROUP BY S.CustomerKey, C.LastName
--OPTION (MAXDOP 1)
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11049
GROUP BY S.CustomerKey, C.LastName
--OPTION (MAXDOP 1)
go

-- Actual tipping point between 301 and 307...

-- Is this still the same in a bigger database on a beefy machine?
-- the concept is the same but the tipping point might be higher
-- up...

USE AdventureWorksDW2008_ModifiedSalesKey
go

SET STATISTICS IO ON
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11065
GROUP BY S.CustomerKey, C.LastName
OPTION (MAXDOP 1)
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11066
GROUP BY S.CustomerKey, C.LastName
OPTION (MAXDOP 1)
go

-- What if we let SQL Server use multiple processors/cores

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11062
GROUP BY S.CustomerKey, C.LastName
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11063
GROUP BY S.CustomerKey, C.LastName
go

-- If this query were critical... 
-- I'd first look at the missing index suggestion 
--   (the "green" suggestion in showplan)

--CREATE NONCLUSTERED INDEX [<Name of Missing Index, sysname,>]
--ON [dbo].[FactInternetSales] ([CustomerKey])
--INCLUDE ([SalesAmount])

-- But... before you create it make sure to FIRST check existing indexes:
-- You can grab sp_SQLskills_SQL2008_helpindex here: 

sp_SQLskills_SQL2008_helpindex FactInternetSales
go

--1) Create the new index

CREATE NONCLUSTERED INDEX IX_FactInternetSales_CustomerKey_INCSalesAmount
ON [dbo].[FactInternetSales] ([CustomerKey])
INCLUDE ([SalesAmount])
go

--2) Drop the old???
--   we should always check our indexes first to see if any can be dropped/consolidated 
--   based on the new index being created!

DROP INDEX [FactInternetSales].IX_FactInternetSales_CustomerKey
go

-- May need to drop the old before creating the new 
-- (if space is an issue)
-- but this may negatively affect existing users?!

-- Now, no matter what the range and/or number of rows, this index can always be used:
SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11048
GROUP BY S.CustomerKey, C.LastName
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11049
GROUP BY S.CustomerKey, C.LastName
go

-- What about an even bigger range:
SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey > 11049
GROUP BY S.CustomerKey, C.LastName
go

------------------------------------------------------------------
-- Clean up and prep for next execution:
------------------------------------------------------------------
USE AdventureWorksDW2008_ModifiedSalesKey
go

CREATE NONCLUSTERED INDEX IX_FactInternetSales_CustomerKey
ON [dbo].[FactInternetSales] ([CustomerKey])
go

DROP INDEX [FactInternetSales].IX_FactInternetSales_CustomerKey_INCSalesAmount
go

